#!/usr/bin/env python

"""
This script will add Numpy support to XALGLIB (CPython wrapper of ALGLIB)

Tested on: 3.5, 3.6, 3.7
"""

# ================
#  Python modules
# ================
import os, sys
import shutil


# ============
#  Parameters
# ============
name = 'xalglib.py'
modif_tag = '# WARNING: modified wrapper version (fork), supporting Numpy arrays'
__version__ = '1.1'
version_tag = 'Done with script v%s' % __version__


# =============
#  File checks
# =============
orig = name.replace('.py', '__original.py')

def is_modified(filename):
    """
    Check if the given filename has already been modified
    """
    f = open(name, 'r')
    res = False
    for i in xrange(25):
        line = f.readline()
        if line.startswith(modif_tag):
            res = True
            break
    f.close()
    return res

# --- backup ---
if not os.path.isfile(orig):
    # 1. make sure the file has not been modified already
    if is_modified(name):
        print 'The file has already been modified, but the original version has not been found. Stopping.'
        sys.exit(0)
    else:
        shutil.copy2(name, orig)
else:
    if is_modified(name):
        print 'The file has already been modified. Stopping.'
        sys.exit(0)
    else:
        print 'The file is not recognized as modified. Using the original version, and redo the modifications.'
        i = 0
        bckp = name
        while os.path.isfile(bckp):
            bckp = name.replace('.py', '__backup-%d.py' % i)
            i += 1
        print '(previous version saved to: %s)' % bckp
        shutil.copy2(name, bckp)



# ===============
#  Modifications
# ===============
# --- import modification blocks ---
import _blocks

class Modifications(object):
    """
    A simple class to perform the modifications on the ALGLIB CPython wrapper.
    """
    def __init__(self, original_file, filename):
        """
        Read data, and create file
        """
        # --- read original file --
        with open(original_file, 'r') as f:
            self.data = f.readlines()
        self.n = len(self.data)
        self.i = 0

        # --- create modified version ---
        self.f = open(filename, 'w')
        self.count = 0   # modifications counter
        self.remodel = 0

    def copy(self, stop):
        """
        Copy the content of the original file from current position to the line
        starting with *stop* (excluded).
        Updates the current position in original file.
        """
        while not self.data[self.i].startswith(stop):
            print >> self.f, self.data[self.i].rstrip()
            self.i += 1

    def insert(self, block):
        """
        Insert block into the modified file.
        """
        print >> self.f, block

    def skip(self, stop):
        """
        Skip the content of the original file from current position to the line
        starting with *stop* (excluded).
        Updates the current position in original file.
        """
        while not self.data[self.i].startswith(stop):
            self.i += 1

    def update_casting(self):
        """
        Updates the cast calls
        """
        print "\nNow updating the cast calls..."
        while self.i < self.n:
            dst = self.data[self.i].lstrip()
            if dst.startswith("if not is_"):
                print "line %4d matching" % (self.i+1)
                if self.remodel != 1:
                    tmp = []
                    self.remodel = 1
                msg = self.data[self.i+1].rstrip().split("ValueError(")[1].split(")")[0]
                tmp.append(msg)
                self.i += 2
            elif dst.startswith("x_from_list") or dst.startswith("x_from_listlist"):
                if self.remodel == 1:
                    self.remodel = 2
                    nrest = 0
                try:
                    print >> self.f, "".join((self.data[self.i].rstrip()[:-1], ", errmsg=", tmp[nrest], ")"))
                    print "changing line: %4d" % (self.i+1)
                    self.count += 1
                    nrest += 1
                except IndexError, err:
                    print "--> found a lonely 'x_from_' at line: %4d" % (self.i+1)
                    self.remodel = 0
                self.i += 1
            else:
                print >> self.f, self.data[self.i].rstrip()
                self.i += 1
        print '\nFile has been modified. %d changes have been done' % self.count
        self.f.close()


M = Modifications(orig, name)

# --- remove and add blocks of code ---
# 1. add modification tag
M.insert('. '.join((modif_tag, version_tag)))

# 2. add new function and constants
M.copy('class x_complex(')
M.insert(_blocks.func_and_ctes)

# 3. replace 'safe_*' functions
M.copy('def safe_len(')
M.insert(_blocks.safes)
M.skip('def create_real_vector(')

# 4. add 'castto' function and replace 'x_from_*', '*_from_x' and 'copy_x_*' functions
M.copy('def is_bool(')
M.insert(_blocks.conversions)
M.skip('_lib_alglib.x_obj_free_hqrndstate.')


# --- update the cast calls ---
M.update_casting()